"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const status_1 = require("./errors/status");
const index_1 = require("./index");
const methods_1 = require("./methods");
const route_1 = require("./route");
const utils_1 = require("./utils");
class Router {
    constructor() {
        this.routes = [];
        this.middlewares = [];
        this.child = (path, handler, options) => {
            let router;
            let routerMiddlewares;
            if (handler instanceof Router) {
                router = handler;
            }
            else if (Array.isArray(handler)) {
                router = handler[0];
                routerMiddlewares = utils_1.toArray(handler[1]);
            }
            if (!router) {
                router = new Router();
            }
            if (routerMiddlewares) {
                router.middleware(routerMiddlewares);
            }
            this.routes.push(new route_1.Route(undefined, path, router, options));
            return router;
        };
        this.route = (method, path, handler, options) => {
            this.routes.push(new route_1.Route(utils_1.toLowerCases(method), path, handler, { exact: true, ...options }));
            return this;
        };
        this.any = (path, handler, options) => {
            return this.route(undefined, path, handler, options);
        };
        this.middleware = (middleware) => {
            if (middleware) {
                this.middlewares.push(...utils_1.toArray(middleware));
            }
            return this;
        };
        this.handle = async (ctx) => {
            try {
                if (!ctx.path || !ctx.path.startsWith("/")) {
                    ctx.path = "/" + (ctx.path || "");
                }
                const url = ctx.path;
                let route;
                let match;
                const params = {};
                const optionsAllowedMethods = [];
                this.routes.some((testRoute) => {
                    if (testRoute.regex) {
                        match = testRoute.regex.exec(url);
                        if (!match) {
                            return false;
                        }
                        else {
                            // Lots of ifs, but need to be explicit for TypeScript
                            if (testRoute.method) {
                                if (Array.isArray(testRoute.method)) {
                                    optionsAllowedMethods.push(...testRoute.method);
                                }
                                else {
                                    optionsAllowedMethods.push(testRoute.method);
                                }
                            }
                            else {
                                optionsAllowedMethods.push(...methods_1.allMethods);
                            }
                        }
                    }
                    const noMatchingMethod = testRoute.method &&
                        testRoute.method !== ctx.method &&
                        !(Array.isArray(testRoute.method) &&
                            testRoute.method.includes(ctx.method));
                    if (noMatchingMethod) {
                        return false;
                    }
                    if (match) {
                        ctx.path = ctx.path.replace(match[0], "");
                    }
                    route = testRoute;
                    return true;
                });
                if (route && match) {
                    const middlewaresNext = [];
                    async function applyNextMiddlewares() {
                        for (const middelwareNext of middlewaresNext) {
                            await middelwareNext();
                        }
                    }
                    try {
                        for (let index = 1; index < match.length; index++) {
                            const key = route.keys[index - 1];
                            const prop = key.name;
                            const value = utils_1.decodeParam(match[index]);
                            if (value !== undefined && !(prop in params)) {
                                params[prop] = value;
                            }
                        }
                        ctx.params = { ...ctx.params, ...params };
                        for (const middleware of this.middlewares) {
                            const next = await middleware(ctx);
                            if (next && typeof next === "function") {
                                middlewaresNext.push(next);
                            }
                        }
                        await index_1.Routex.runHandler(route.handle, ctx);
                        if (ctx.error) {
                            throw ctx.error;
                        }
                    }
                    catch (error) {
                        await applyNextMiddlewares();
                        throw error;
                    }
                    await applyNextMiddlewares();
                    return;
                }
                if (ctx.method === methods_1.Methods.OPTIONS) {
                    // Prevent duplicates methods, all upper-case
                    const methods = [];
                    optionsAllowedMethods.forEach((optionsMethod) => {
                        if (methods.includes(optionsMethod.toUpperCase())) {
                            return;
                        }
                        methods.push(optionsMethod.toUpperCase());
                    });
                    ctx.res.setHeader("Allow", methods.join(", "));
                    return;
                }
                // 404
                throw new status_1.ErrorWithStatusCode(404, "Not Found");
            }
            catch (error) {
                if (this.errorHandler) {
                    return this.errorHandler(ctx, error);
                }
                throw error;
            }
        };
    }
}
exports.Router = Router;
// Attach method shortcuts such as router.get(...), etc
Object.values(methods_1.Methods).forEach((method) => {
    // eslint-disable-next-line @typescript-eslint/ban-ts-ignore
    // @ts-ignore
    Router.prototype[method] = function (path, handler, options) {
        return this.route(method, path, handler, options);
    };
});
//# sourceMappingURL=router.js.map