"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const http = require("http");
const https = require("https");
const parseUrl = require("parseurl");
const qs = require("qs");
const throng = require("throng");
const uuid_1 = require("uuid");
const defaultHandler_1 = require("./errors/defaultHandler");
const methods_1 = require("./methods");
const router_1 = require("./router");
const utils_1 = require("./utils");
class Routex extends router_1.Router {
    constructor({ requestId, errorHandler, providers, } = {}) {
        super();
        this.errorHandler = defaultHandler_1.defaultErrorHandler;
        this.appMiddlewares = [];
        /**
         * Adds app middleware(s)
         */
        this.appMiddleware = (appMiddleware) => {
            if (appMiddleware) {
                this.appMiddlewares.push(...utils_1.toArray(appMiddleware).map((appMiddlewareOne) => appMiddlewareOne(this)));
            }
            return this;
        };
        /**
         * Handles HTTP request
         */
        this.handler = async (req, res) => {
            // Parse query string and extract path
            const parsed = parseUrl(req);
            const query = parsed &&
                (utils_1.isString(parsed.query)
                    ? qs.parse(parsed.query)
                    : parsed.query);
            const ctx = Routex.createCtx({
                path: (parsed && parsed.pathname) || "/",
                query,
                method: req.method,
                requestId: this.requestIdGenerator
                    ? this.requestIdGenerator()
                    : undefined,
                workerId: this.workerId,
                providers: this.providers,
                req: req,
                res,
            });
            await this.handle(ctx);
            const chunk = ctx.body ? ctx.body.toBuffer() : undefined;
            if (ctx.body) {
                if (ctx.body.contentType) {
                    ctx.res.setHeader("Content-Type", ctx.body.contentType);
                }
                if (chunk && (!ctx.statusCode || ![204, 304].includes(ctx.statusCode))) {
                    ctx.res.setHeader("Content-Length", chunk.length);
                }
            }
            if (ctx.statusCode) {
                ctx.res.writeHead(ctx.statusCode);
            }
            if (chunk && ctx.body && ctx.method !== methods_1.Methods.HEAD) {
                ctx.res.end(chunk);
            }
            else {
                ctx.res.end();
            }
            return ctx;
        };
        if (requestId !== false) {
            this.requestIdGenerator = requestId || (() => uuid_1.v4());
        }
        if (errorHandler) {
            this.errorHandler = errorHandler;
        }
        this.providers = providers || {};
    }
    /**
     * Starts the server
     */
    async listen(port, { hostname, server: optionsServer, https: httpsOptions, } = {}) {
        if (port) {
            const originalPort = port;
            if (utils_1.isString(port)) {
                port = parseInt(port, 10);
            }
            if (!Number.isInteger(port)) {
                throw new TypeError(`Invalid port (${originalPort} is not integer/string).`);
            }
        }
        if (hostname && !utils_1.isString(hostname)) {
            throw new TypeError(`Invalid hostname (${hostname} is not a string)`);
        }
        const server = this.initializeServer(optionsServer ||
            (httpsOptions
                ? https.createServer(httpsOptions, this.handler)
                : http.createServer(this.handler)));
        await new Promise((resolve) => {
            server.listen({ port: port, host: hostname }, resolve);
        });
        const close = () => new Promise((resolve, reject) => {
            server.close((error) => {
                /* istanbul ignore next */
                if (error) {
                    reject(error);
                }
                resolve();
            });
        });
        const address = server.address();
        return {
            close,
            port: 
            /* istanbul ignore next */
            address && !utils_1.isString(address) ? address.port : null,
            server,
        };
    }
    /* istanbul ignore next */
    cluster(port, { workers, ...options } = {}) {
        throng({
            start: (id) => {
                this.workerId = id;
                this.listen(port, options);
            },
            workers,
        });
    }
    /**
     * Decorates a server using app middlewares
     */
    initializeServer(server) {
        return this.appMiddlewares.reduce((reducingServer, appMiddleware) => (appMiddleware.initializeServer &&
            appMiddleware.initializeServer(reducingServer)) ||
            reducingServer, server);
    }
    /**
     * Run a handler, with associated "magic" logic
     */
    static async runHandler(handler, ctx) {
        try {
            const body = await handler(ctx);
            if (body) {
                ctx.body = body;
            }
        }
        catch (error) {
            ctx.error = error;
        }
        return ctx;
    }
}
exports.Routex = Routex;
/**
 * Create a context object. Useful during testing
 */
Routex.createCtx = ({ path, query, req, res, requestId, workerId, providers, method, }) => {
    /* istanbul ignore next */
    return {
        data: {},
        params: {},
        // Default to GET for missing methods (should never happen)
        method: (method || methods_1.Methods.GET).toLowerCase(),
        path: path || "/",
        query: query || {},
        req,
        requestId,
        res,
        workerId,
        providers: { ...providers },
    };
};
//# sourceMappingURL=routex.js.map